@@ -2,8 +2,8 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2009  Intel Corporation. All rights reserved.
- *  Copyright (C) 2010 ST-Ericsson AB.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2010  ST-Ericsson AB.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -313,6 +313,15 @@ static void esimsr_notify(GAtResult *result, gpointer user_data)
 	handle_sim_status(status, modem);
 }
 
+static void epev_notify(GAtResult *result, gpointer user_data)
+{
+	struct ofono_modem *modem = user_data;
+	struct ofono_sim *sim = ofono_modem_get_sim(modem);
+
+	if (sim)
+		ofono_sim_initialized_notify(sim);
+}
+
 static int ste_enable(struct ofono_modem *modem)
 {
 	struct ste_data *data = ofono_modem_get_data(modem);
@@ -353,6 +362,9 @@ static int ste_enable(struct ofono_modem *modem)
 	g_at_chat_register(data->chat[AT_SIM], "*ESIMSR:", esimsr_notify,
 				FALSE, modem, NULL);
 
+	g_at_chat_send(data->chat[AT_SIM], "AT*EPEE=1", NULL, NULL, NULL, NULL);
+	g_at_chat_register(data->chat[AT_SIM], "*EPEV", epev_notify,
+					FALSE, modem, NULL);
 	return -EINPROGRESS;
 
 error:
@@ -423,9 +435,8 @@ static void ste_set_online(struct ofono_modem *modem, ofono_bool_t online,
 	if (g_at_chat_send(chat, command, NULL, set_online_cb, cbd, g_free))
 		return;
 
-	g_free(cbd);
-
 	CALLBACK_WITH_FAILURE(cb, cbd->data);
+	g_free(cbd);
 }
 
 static void ste_pre_sim(struct ofono_modem *modem)
@@ -501,10 +512,10 @@ static struct ofono_modem_driver ste_driver = {
 	.remove		= ste_remove,
 	.enable		= ste_enable,
 	.disable	= ste_disable,
-	.set_online     = ste_set_online,
+	.set_online	= ste_set_online,
 	.pre_sim	= ste_pre_sim,
 	.post_sim	= ste_post_sim,
-	.post_online    = ste_post_online,
+	.post_online	= ste_post_online,
 };
 
 static int ste_init(void)
